
/* Network traffic analyzer
 * (c) 2001, Emil Mikulic.
 */

/* This is the DNS resolver */

#include "dns.h"
#include "host_db.h"
#include "bignum.h"
#include <netdb.h>
#include <netinet/in.h>
#include <assert.h>
#include <unistd.h>
#include <signal.h>

#define WAIT_BETWEEN_QUERIES 1000*MSEC
#define WAIT_BETWEEN_LOOPS 30*1000*MSEC

pthread_t dns_thread;
pthread_mutex_t db_mutex;



char *name_from_ip(unsigned int ip_addr)
{
	struct hostent *hp;
	unsigned int ip;

	ip = htonl(ip_addr);
	hp = gethostbyaddr((char*)&ip, 4, AF_INET);

	if (hp)
		return copy_str(hp->h_name);
	else
		return NULL;
}



/* argument is ignored */
void dns_main(void *ptr)
{
	int i = 0;
	sigset_t sigset;
	
	sigfillset(&sigset);
	pthread_sigmask(SIG_BLOCK, &sigset, NULL);

	printf("DNS: Thread is awake. (PID:%d)\n", getpid());
	fflush(stdout);

	while (host_db.used < 1)
	{
		printf("DNS: Waiting for the host_db to fill up.\n");
		fflush(stdout);
		usleep(WAIT_BETWEEN_QUERIES);
	}

	while (1)
	{
		host_record *h;
		int db_used;
		
		pthread_mutex_lock(&db_mutex);
		h = (host_record*)db_get(host_db, i);
		db_used = host_db.used;
		pthread_mutex_unlock(&db_mutex);

		if (h->hostname == NULL)
		{
			h->hostname = name_from_ip(h->ip_addr);
			if (h->hostname)
				printf("DNS: Resolved %s", h->hostname);
			else
				printf("DNS: Resolution failed");
			printf(" (%d/%d)\n", i+1, db_used);
			fflush(stdout);
			usleep(WAIT_BETWEEN_QUERIES);
		}

		i++;
		if (i == db_used)
		{
			i = 0;

			printf("DNS: Hit end of list -> sleeptime\n");
			fflush(stdout);

			usleep(WAIT_BETWEEN_LOOPS);
		}
	}

	assert(0 && "Execution should never reach this point.");
	pthread_exit(0);
}

