
/* Network traffic analyzer
 * (c) 2001, Emil Mikulic.
 */

/* Port database */

#include "port_db.h"



/* Global database */
db port_db;



void port_db_init(void)
{
	db_init(&port_db);
}

void port_db_free(void)
{
	unsigned int i;

	for (i=0; i<port_db.used; i++)
		free(port_db.records[i]);

	db_free(&port_db);
}



/* port_record needs to be deallocated when it's no longer needed */
port_record *port_make(const word port)
{
	port_record *p = (port_record*)malloc(sizeof(port_record));

	p->port = port;
	p->data_in.hi = p->data_in.lo =
		p->data_out.hi = p->data_out.lo = 0;

	return p;
}



/* update transfer amount */
__inline void port_update_in(port_record *p, const dword in)
{
	i64 add = {0, in};
	add_64bit(&(p->data_in), add);
}

__inline void port_update_out(port_record *p, const dword out)
{
	i64 add = {0, out};
	add_64bit(&(p->data_out), add);
}



/* find a port */
port_record *port_find(const word port)
{
	int i;

	if (port_db.used == 0)
		return NULL;
	
	for (i=0; i<port_db.used; i++)
	{
		port_record *tmp = (port_record*)db_get(port_db,i);
		assert(tmp != NULL);
		if (tmp->port == port) return tmp;
	}

	return NULL;
}



/* add or update a port, depending on whether it exists in the db or not */
__inline port_record *port_from_num(const word port)
{
	port_record *p;
	
	p = port_find(port);

	if (p == NULL)
	{
		p = port_make(port);
		if (!db_add(&port_db, p)) freakout();
	}

	return p;
}



void port_db_dump(FILE *fp)
{
	int i;

	fprintf(fp, "%u\n", port_db.used);
	for (i=0; i<port_db.used; i++)
	{
		port_record *p = (port_record*)db_get(port_db, i);
		
		fprintf(fp, "%u|%u:%u|%u:%u\n",
			p->port,
			p->data_in.hi,
			p->data_in.lo,
			p->data_out.hi,
			p->data_out.lo);
	}
}

