
/* Network traffic analyzer
 * (c) 2001, Emil Mikulic.
 */

/* Horribly incorrect and minimalistic web server */

#include <string.h>
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/times.h>
#include <sys/wait.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <signal.h>
#include <stdarg.h>

#include "version.h"
#include "www.h"
#include "host_db.h"
#include "port_db.h"
#include "proto_db.h"
#include "bignum.h"
#include "graph.h"
#include "gif.h"

#define WEBPORT 666
#define POLLDELAY 5 /* seconds */
#define MAXCHILDREN 5 /* won't fire up more threads than this */
#define SOCKET int /* WinSock? err... */

pthread_t www_thread;



/* Handling fatal errors */
#define error(s) { fprintf(stderr, "WWW: " s "\n"); exit(101); }



static char www_unimplemented[] = "<html><body>\n<h1>Not Implemented</h1>\nWhatever the heck "
	"you just requested, I can't generate.\n</body></html>\n";

static char www_not_found[] = "<html><body>\n<h1>Not Found</h1>\nWhatever you just requested, "
	"I don't have.\n</body></html>\n";



static char www_content_html[] = "Content-type: text/html\r\n";
static char www_content_gif[] = "Content-type: image/gif\r\n";
static char www_noproxy[] = "Cache-Control: no-cache\r\n"
	"Pragma: no-cache\r\n";
static char www_connection[] = "Connection: close\r\n";
static char www_server[] = "Server: darkstat\r\n";
static char www_end_header[] = "\r\n";



void appendf(char **str, char *format, ...)
{
	int len1, len2;
	char buf[1024];
	va_list va;
	
	va_start(va, format);
	vsprintf(buf, format, va);
	va_end(va);

	if ((*str) != NULL) len1 = strlen(*str); else len1 = 0;
	len2 = strlen(buf)+len1+1;
	
	*str = realloc(*str, len2);
	if (*str == NULL) error("realloc() failed in appendf()");

	if (len1 == 0) strcpy(*str, buf); else
		strcat(*str, buf);
}



char *www_frontpage(SOCKET incoming)
{
	char *content = NULL, *x_packets, *x_data;
	extern i64 num_packets, total_data;
	extern time_t t_start, t_already;

	time_t uptime = t_already + (time(NULL) - t_start);
	dword days, hours, minutes;
	dword max_transfer = 0;
	int i;
	double divisor;
	
	appendf(&content, "<html><body><h1>" PACKAGE " v" VERSION "</h1>\n");

	/* packets */
	x_packets = bignum_from_64bit(num_packets);
	appendf(&content, "<b>Packets captured:</b> %s<br>\n", x_packets);
	free(x_packets);

	/* data */
	x_data = bignum_from_64bit(total_data);
	appendf(&content, "<b>Data off the wire:</b> %s<br>\n", x_data);
	free(x_data);

#define S_MIN 60
#define S_HR (S_MIN*60)
#define S_DAY (S_HR*24)

	days = uptime / S_DAY;
	uptime -= days * S_DAY;

	hours = uptime / S_HR;
	uptime -= hours * S_HR;

	minutes = uptime / S_MIN;
	uptime -= minutes * S_MIN;

	appendf(&content, "<b>Counting:</b> ");
	if (days) appendf(&content, "%u days ", days);
	if (days || hours) appendf(&content, "%u hours ", hours);
	if (days || hours || minutes) appendf(&content, "%u minutes ", minutes);
	appendf(&content, "%u seconds<br>\n", uptime);

	appendf(&content, "<br>\n");
	appendf(&content, "<table border=0 cellspacing=0 cellpadding=1><tr><td bgcolor=#000000>\n");
	appendf(&content, "<table border=0 cellspacing=0 cellpadding=1><tr><td align=center>");

	pthread_mutex_lock(&graph_mutex);
	for (i=0; i<GRAPH_SECONDS; i++)
		if (i != pos_sec)
			max_transfer = max(max_transfer,
					max(graph_sec_in[i],graph_sec_out[i]));

	appendf(&content, "<font color=\"#FFFFFF\"><b>Last 60 seconds</b> ");
	appendf(&content, "(<font color=#FF0000>in</font> <font color=#FF00FF>and</font> <font color=#0000FF>out</font>)<br>\n");
	appendf(&content, "Maximum: %u</td></tr>\n", max_transfer);
	appendf(&content, "<tr><td bgcolor=\"#FFFFFF\">\n");
	appendf(&content, "<table border=\"0\" cellspacing=\"0\" cellpadding=\"1\"><tr>\n");

#define GRAPHWIDTH "width=5"
#define GRAPHHEIGHT 200

	divisor = max(1.0, (double)max_transfer / (double)GRAPHHEIGHT);

#define GRAPHLINE(i) { 					\
	int in = (int)((double)graph_sec_in[i] / divisor);		\
	int out = (int)((double)graph_sec_out[i] / divisor);		\
	int blank = GRAPHHEIGHT - max(in, out);		\
	int shared = min(in, out);			\
	appendf(&content, "<td valign=bottom>");	\
	if (blank) appendf(&content, "<img src=i/blank " GRAPHWIDTH " height=%d><br>", blank); \
	if ((in > out) && (in-shared))			\
		appendf(&content, "<img src=i/r " GRAPHWIDTH " height=%d><br>", in-shared); \
	else  if (out-shared) 				\
		appendf(&content, "<img src=i/b " GRAPHWIDTH " height=%d><br>", out-shared); \
	if (shared) appendf(&content, "<img src=i/m " GRAPHWIDTH " height=%d>", shared); \
	appendf(&content, "</td>\n"); }

	for (i=pos_sec+1; i<=GRAPH_SECONDS; i++) GRAPHLINE(i);
	for (i=0; i<pos_sec; i++) GRAPHLINE(i);

	pthread_mutex_unlock(&graph_mutex);
#undef GRAPHLINE
	appendf(&content, "</tr></table>\n");
	appendf(&content, "</td></tr></table></td></tr></table>\n");

	appendf(&content, "<br>\n");
	appendf(&content, "[<a href=\"hosts.html\">hosts</a>]\n");
	appendf(&content, "[<a href=\"ports.html\">ports</a>]\n");
	appendf(&content, "[<a href=\"protocols.html\">protocols</a>]\n");
	appendf(&content, "<br>\n</body></html>\n");

	return content;
}



char *www_hosts(SOCKET incoming)
{
	int i;
	char *buf = NULL;

	appendf(&buf, "<html><body><h1>Hosts</h1>\n");
	appendf(&buf, "<a href=\"/\">&lt;&lt;</a> home.<br><br>\n");
	appendf(&buf, "<table border=\"0\" cellpadding=\"1\" cellspacing=\"0\"><tr><td bgcolor=\"#000000\">\n");
	appendf(&buf, "<table border=\"0\" cellspacing=\"1\" cellpadding=\"5\">");
	appendf(&buf, "<tr><td bgcolor=\"#C0C0C0\">IP</td><td bgcolor=\"#C0C0C0\">Hostname</td><td bgcolor=\"#C0C0C0\">Data in</td><td bgcolor=\"#C0C0C0\">Data out</td><td bgcolor=\"#C0C0C0\">Total</td></tr>\n");

	pthread_mutex_lock(&db_mutex);

	for (i=0; i<host_db.used; i++)
	{
		host_record *h;
		char *data_in, *data_out, *ip_quad, *hostname, *total;

		/* grab the record */	
		h = (host_record*)db_get(host_db, i);

		/* ip */
		ip_quad = quad_from_ip(h->ip_addr);
		
		/* hostname */
		hostname = (h->hostname)?h->hostname:"&nbsp;";

		/* data in */
		data_in = bignum_from_64bit(h->data_in);

		/* data out */
		data_out = bignum_from_64bit(h->data_out);

		total = add_bignum(data_in, data_out);

		appendf(&buf,"<tr><td bgcolor=\"#FFFFFF\">%s</td><td bgcolor=\"#FFFFFF\">%s</td><td bgcolor=\"#FFFFFF\">%s</td><td bgcolor=\"#FFFFFF\">%s</td><td bgcolor=\"#FFFFFF\">%s</td></tr>\n",
				ip_quad, hostname, data_in, data_out, total);

		free(data_in);
		free(data_out);
		free(ip_quad);
		free(total);
	}
	pthread_mutex_unlock(&db_mutex);

	appendf(&buf, "</table>\n</td></tr></table>\n</body></html>\n");

	return buf;
}



void port_rundown(char **buf, SOCKET incoming, char *type_name, db port_db)
{
	int i;
	
	appendf(buf, "<table border=\"0\" cellpadding=\"1\" cellspacing=\"0\"><tr><td bgcolor=\"#000000\">\n");
	appendf(buf, "<table border=\"0\" cellspacing=\"1\" cellpadding=\"5\">");

	appendf(buf,"<tr><td bgcolor=\"#000000\" colspan=\"4\" align=\"center\"><font color=\"#FFFFFF\"><b>%s</b></font></td></tr>\n", type_name);

	appendf(buf, "<tr><td bgcolor=\"#C0C0C0\">Port</td><td bgcolor=\"#C0C0C0\">Data in</td><td bgcolor=\"#C0C0C0\">Data out</td><td bgcolor=\"#C0C0C0\">Total</td></tr>\n");

	pthread_mutex_lock(&db_mutex);

	for (i=0; i<port_db.used; i++)
	{
		port_record *p;
		char *data_in, *data_out, *total;

		/* grab the record */	
		p = (port_record*)db_get(port_db, i);

		/* data in */
		data_in = bignum_from_64bit(p->data_in);

		/* data out */
		data_out = bignum_from_64bit(p->data_out);

		total = add_bignum(data_in, data_out);

		appendf(buf,"<tr><td bgcolor=\"#FFFFFF\">%d</td><td bgcolor=\"#FFFFFF\">%s</td><td bgcolor=\"#FFFFFF\">%s</td><td bgcolor=\"#FFFFFF\">%s</td></tr>\n",
				p->port, data_in, data_out, total);

		free(data_in);
		free(data_out);
		free(total);
	}
	pthread_mutex_unlock(&db_mutex);

	appendf(buf, "</table>\n</td></tr></table>\n");
}

char *www_ports(SOCKET incoming)
{
	char *buf = NULL;

	appendf(&buf, "<html><body><h1>Ports</h1>\n");
	appendf(&buf, "<a href=\"/\">&lt;&lt;</a> home.<br><br>\n");

	port_rundown(&buf, incoming, "Transmission Control Protocol", port_db);
	appendf(&buf, "</body></html>\n");

	return buf;
}



char *www_protocols(SOCKET incoming)
{
	int i;
	char *buf = NULL;

	appendf(&buf, "<html><body><h1>Protocols</h1>\n");
	appendf(&buf, "<a href=\"/\">&lt;&lt;</a> home.<br><br>\n");
	
	appendf(&buf, "<table border=\"0\" cellpadding=\"1\" cellspacing=\"0\"><tr><td bgcolor=\"#000000\">\n");
	appendf(&buf, "<table border=\"0\" cellspacing=\"1\" cellpadding=\"5\">");
	appendf(&buf, "<tr><td colspan=\"2\" bgcolor=\"#C0C0C0\">Protocol</td><td bgcolor=\"#C0C0C0\">Data in</td><td bgcolor=\"#C0C0C0\">Data out</td><td bgcolor=\"#C0C0C0\">Total</td></tr>\n");

	pthread_mutex_lock(&db_mutex);

	for (i=0; i<proto_db.used; i++)
	{
		proto_record *p;
		char *data_in, *data_out, *total;

		/* grab the record */	
		p = (proto_record*)db_get(proto_db, i);

		/* data in */
		data_in = bignum_from_64bit(p->data_in);

		/* data out */
		data_out = bignum_from_64bit(p->data_out);

		total = add_bignum(data_in, data_out);

		appendf(&buf, "<tr><td bgcolor=\"#FFFFFF\" align=\"right\">%d</td><td bgcolor=\"#FFFFFF\">%s</td><td bgcolor=\"#FFFFFF\">%s</td><td bgcolor=\"#FFFFFF\">%s</td><td bgcolor=\"#FFFFFF\">%s</td></tr>\n",
			p->protocol, proto_name(p->protocol), data_in, data_out, total);

		free(data_in);
		free(data_out);
		free(total);
	}
	pthread_mutex_unlock(&db_mutex);

	appendf(&buf, "</table>\n</td></tr></table>\n");
	appendf(&buf, "</body></html>\n");

	return buf;
}



char *new_gif(void)
{
	char *buf = (char*)malloc(GIF_SIZE);
	memcpy(buf, gif, GIF_SIZE);
	return buf;
}



void handle_connection(SOCKET incoming)
{
	char buf[4096], *hdr = NULL, *content = NULL;
	struct linger lng;
	int recvd, is_gif = 0;

	recvd = recv(incoming, buf, 4096, 0);
	buf[recvd] = 0;

#define SEND(x) send(incoming,x,strlen(x),0)
#define CMP(x) (memcmp(buf,x,strlen(x)) == 0)

	if (!CMP("GET ")) content = www_unimplemented; else
	if (CMP("GET / ") || CMP("GET /index.html "))
		content = www_frontpage(incoming); else
	if CMP("GET /hosts.html ") content = www_hosts(incoming); else
	if CMP("GET /ports.html ") content = www_ports(incoming); else
	if CMP("GET /protocols.html ") content = www_protocols(incoming); else
	if CMP("GET /i/")
	{
		content = new_gif();
		is_gif = 1;


		if CMP("GET /i/r ") content[GIF_RED] = 0xFF; else
		if CMP("GET /i/b ") content[GIF_BLUE] = 0xFF; else
		if CMP("GET /i/m ") content[GIF_RED] = content[GIF_BLUE] = 0xFF; else
		if CMP("GET /i/blank ") content[GIF_PIXEL] = GIF_TRANSPARENT; else
		{
			free(content);
			is_gif = 0;
			content = www_not_found;
		}
	} else
		content = www_not_found;

	if (content == www_unimplemented) appendf(&hdr, "HTTP/1.0 501 Not Implemented\r\n"); else
	if (content == www_not_found) appendf(&hdr, "HTTP/1.0 404 Not Found\r\n"); else
		appendf(&hdr, "HTTP/1.0 200 OK\r\n");

	appendf(&hdr, www_server);
	if (!is_gif) appendf(&hdr, www_noproxy);

	appendf(&hdr, "Accept-Ranges: bytes\r\n");
	appendf(&hdr, "Content-Length: %d\r\n", (is_gif) ? GIF_SIZE : strlen(content));
	appendf(&hdr, www_connection);
	if (is_gif)
		appendf(&hdr, www_content_gif);
	else
		appendf(&hdr, www_content_html);
	appendf(&hdr, www_end_header);
	SEND(hdr);
	free(hdr);

	if (is_gif)
		send(incoming, content, GIF_SIZE, 0);
	else
		SEND(content);

	if ( (content != www_unimplemented) && (content != www_not_found) )
			free(content);

#undef CMP
#undef SENT

	/* stupid lingering */
	lng.l_onoff = 1;
	lng.l_linger = 1;
	setsockopt(incoming, SOL_SOCKET, SO_LINGER, &lng, sizeof(lng));

	/* die die die! */
	shutdown(incoming, 2);
	close(incoming);
	shutdown(incoming, 2);
}



SOCKET sockin;

void www_main(void *ptr)
{
	struct sockaddr_in addrin;
	sigset_t sigset;
        
	sigfillset(&sigset);
	pthread_sigmask(SIG_BLOCK, &sigset, NULL);

	/* create the incoming socket */
	sockin = socket(AF_INET, SOCK_STREAM, 0);
	if (sockin < 0) error("Problem creating socket");

	/* fill out a sockaddr struct */
	addrin.sin_family = AF_INET;
	addrin.sin_port = htons(WEBPORT);
	addrin.sin_addr.s_addr = INADDR_ANY;
	bzero(&(addrin.sin_zero), 8);

	/* bind sockin to the incoming port */
	if (bind(sockin, (struct sockaddr *)&addrin,
				sizeof(struct sockaddr)) < 0)
		error("Problem binding socket");

	/* listen on the socket */
	if (listen(sockin, MAXCHILDREN) < 0)
		error("Problem listen()ing to socket");

	printf("WWW: Thread is awake and waiting for connections. (PID:%d)\n", getpid());
	while (1)
	{
		SOCKET incoming;
		int sin_size, select_ret;
		struct timeval timeout;
		fd_set acceptfds;
		int wait_status, pid;

		/* wait for a connection */
		FD_ZERO(&acceptfds);
		FD_SET(sockin, &acceptfds);

		timeout.tv_sec = POLLDELAY;
		timeout.tv_usec = 0;

		select_ret = select(sockin+1, &acceptfds,
			NULL, NULL, &timeout);

		if (select_ret == -1) error("select() died on me");
		if (select_ret == 1) {
			int fork_ret;
			
			sin_size = sizeof(struct sockaddr);
			incoming = accept(sockin, (struct sockaddr *)&addrin,
				&sin_size);
			if (incoming < 0)
				error("Problem accept()ing connection");

			/* details about incoming connection */
			printf("WWW: Got a connection from %s:%d\n",
				inet_ntoa(addrin.sin_addr),
				ntohs(addrin.sin_port));
			fflush(stdout);

			/* fork another process for this connection */
			fork_ret = fork();
			if (fork_ret == 0) {
				handle_connection(incoming);
				exit(0);
			} else if (fork_ret == -1) {
				printf("WWW: Unable to fork()\n");
				close(incoming);
			} else {
				printf("WWW: Forked child process (PID:%d)\n", fork_ret);
			}
		}
		
		/* see if any children are finished */
		do {
			pid = waitpid(-1, &wait_status, WNOHANG);
			if (pid > 0)
				printf("WWW: process (PID:%d) terminated\n", pid);
		} while (pid>0);

		fflush(stdout);
	}

	assert(0 && "Execution should never reach this point.");
	pthread_exit(0);
}



void www_shutdown(void)
{
	shutdown(sockin, 2);
	close(sockin);
}

